import { State } from './state.js'
import { anyOf } from './combinator.js'
import { Alpha, Alphanumeric, Digits, LowerAlpha, UpperAlpha } from './const.js'
import { fail, join, mapStr, next, succeed } from './fn.js'
import { map, seq } from './seq.js'
import { curry } from '../vendor/izuna/src/curry.js'

/** @import { ParserState } from './state.js' */

export const char = curry(
  /**
   * @param {string} ch
   * @param {ParserState} state
   */
  (ch, state) => (
    next(state) === ch ? succeed(ch, state) : fail(`could not parse ${ch} `, state)
  ))

export const str = curry(
  /**
   * @param {string} str
   * @param {State} state
   */
  (str, state) => (
    map(
      join(''),
      seq(...mapStr(char, str))
    )(state)
  ))

export const anyChar = state => {
  const ch = next(state)
  return !!ch ? succeed(ch, state) : fail('end of input', state)
}

export const digit = anyOf(Digits)
export const lowerAlpha = anyOf(LowerAlpha)
export const upperAlpha = anyOf(UpperAlpha)
export const alpha = anyOf(Alpha)
export const alphanumeric = anyOf(Alphanumeric)

