import { ParseError } from './state.js'
import { fork, succeed } from './fn.js'
import { curry } from '../vendor/izuna/src/index.js'
import { anyChar } from './char.js'

/** @import { Result } from '../vendor/kojima/src/index.js' */
/** @import { ParserState } from './state.js' */

export const maybe = curry(
  /**
    * @param {(...args: any[]) => Result<ParserState, ParseError>} parser
    * @param {ParserState} state
    */
  (parser, state) => {
    const [original, clone] = fork(state)
    const result = parser(clone)
    return result.isOk() ? result : succeed([], original)
  })

export const not = parser => state => {
  const result = parser(state)
  if (result.isOk()) {
    return fail('"not" parser failed', state)
  } else {
    return succeed([], state)
  }
}

export const until = parser => state => {
  let acc = ok(state)

  while (result.isOk()) {
    const [original, clone] = fork(state)
    const result = acc.chain(x => parser(clone))
    if (result.isOk()) {
      break
    } else {
      acc = anyChar(original)
    }
  }

  return acc
}

export const many = curry((parser, state) => {
  let result = ok(state)

  while (true) {
    const [original, clone] = fork(result)
    const res = result.chain(x => parser(clone))
    if (res.isOk()) {
      result = res
    } else {
      break
    }
  }

  return result
})

