import { ParseError } from './state.js'
import { Iter } from './iter.js'
import { err, ok } from '../vendor/kojima/src/index.js'
import { curry } from '../vendor/izuna/src/index.js'

/** @import { ParserState } from './state.js'* /

/**
 * @param {...any} values
 */
export const Tuple = (...values) => Object.freeze(values)

/**
 * @template T
 * @param {Iterator<T> | Iterable<T>} iterable
 * @param {number} [n=2]
 */
function tee(iterable, n = 2) {
  const iterator = Iter.from(iterable)

  /**
   * @param {{ next: any, value: T }} current 
   */
  function* gen(current) {
    while (true) {
      if (!current.next) {
        const { done, value } = iterator.next()
        if (done) return
        current.next = { value }
      }

      current = current.next
      yield current.value
    }
  }

  return Array(n).fill({}).map(gen)
}

/**
 * @param {import('./state.js').ParserState} state
 */
export const fork = ([tokens, state]) => {
  const [a, b] = tee(state)
  return Tuple(
    Tuple(tokens.slice(), a),
    Tuple(tokens.slice(), b),
  )
}

/**
 * @template T
 * @param {T | T[]} v
 * @param {ParserState} state
 */
export const succeed = (v, [x, y]) => ok(Tuple(x.concat(v), y))

/**
 * @param {string} msg
 * @param {ParserState} state
 * @param {Error} [e]
 */
export const fail = (msg, state, e = undefined) => err(new ParseError(msg, state, e))

/**
 * @template T
 * @param {number & keyof T} n
 * @param {T[]} iter
 */
export const nth = (n, iter) => iter[n]

/**
 * @param {ParserState} state
 */
export const next = state => state[1].next().value

/**
 * @template T
 * @param {T[]} a
 * @param {T[]} b
 */
export const diff = (a, b) => b.slice(-Math.max(0, b.length - a.length))

export const join = curry(
  /**
    * @param {string} delim
    * @param {string[]} val
    */
  (delim, val) => val.join(delim)
)

export const mapStr = curry(
  /**
    * @param {(...args: any[]) => any} fn
    * @param {string} str
    */
  (fn, str) => Array.from(str).map(v => fn(v))
)

